#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" basic container types """

from ujotypes import UjoList, UjoMap

from .basic import UjsBase


class UjsList(UjsBase):
    name = "list"
    typeid = 0x30
    default_to_ujo = UjoList

    def __init__(self, typ):
        self.type = typ

    def __str__(self):
        return self.name + " of " + repr(self.type)

    def __repr__(self):
        return self.name + " of " + repr(self.type)


class UjsMap(UjsBase):
    name = "map"
    typeid = 0x31
    default_to_ujo = UjoMap

    def __init__(self, typ):
        self.type = typ

    def __str__(self):
        return self.name + " of " + repr(self.type)

    def __repr__(self):
        return self.name + " of " + repr(self.type)



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

# flake8: noqa

from .binary import schema_to_binary, SchemaToUjo
from .markdown import schema_to_markdown, SchemaToMarkdown
from .json import schema_to_json, SchemaToJSON
from .ujotypes import schema_to_type, SchemaToType



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" conversion of a schema to Ujo binary representation """
from abc import abstractmethod, ABC
from copy import deepcopy

from ujotypes import (
    UJO_VARIANT_NONE,
    UjoUInt8,
    UjoList,
    UjoStringUTF8,
    UjoUInt32,
    UjoMap,
    UjoBool,
)

from UJOSchema.types import (
    Module,
    Typedef,
    CustomType,
    UjsList,
    UjsMap,
    Object,
    Record,
    RecordItem,
)
from UJOSchema.constraints import UjsNotNull, UjsIn, UjsNotIn, UjsOfLength
from UJOSchema.ranges import UjsEquals, UjsBetween, UjsAbove, UjsWord, UjsBelow

from .base import (
    BaseSchemaConverter,
    BaseTypeConverter,
    BaseConstraintConverter,
    BaseRangeConverter,
)


def schema_to_binary(schema, target=None):
    return bytes(SchemaToUjo(schema, target).convert())


class SchemaToUjo(BaseSchemaConverter):
    def convert(self):
        if self.target:
            return TypeToUjoConverter(self.schema, self.schema[self.target]).convert()

        converted = [
            TypeToUjoConverter(self.schema, typ).convert()
            for typ in self.schema.values()
            if not isinstance(typ, Module)
        ]

        if len(converted) == 1:
            return next(iter(converted))

        results = UjoList()
        for typ in converted:
            results.append(typ)
        return results


class TypeToUjoConverter(BaseTypeConverter):
    @property
    def type_name(self):
        return UjoStringUTF8(self.type.name)

    @property
    def type_id(self):
        return UjoUInt8(self.type.typeid)

    def convert(self):
        return self.convert_type(self.type)

    def create_container(self):
        container = UjoList()
        container.append(self.type_name)
        container.append(self.type_id)
        return container

    def convert_type(self, typ):
        default_converter = BasicTypeToUjo
        converters = {
            list: ListOfTypesToUjo,
            tuple: ListOfTypesToUjo,
            Typedef: TypedefToUjo,
            UjsList: ContainerToUjo,
            UjsMap: ContainerToUjo,
            CustomType: CustomTypeToUjo,
            Record: RecordToUjo,
            RecordItem: TypedefToUjo,
            Object: ObjectToUjo,
        }

        converter = converters.get(type(typ)) or default_converter
        return converter(self.schema, typ).convert()


class BasicTypeToUjo(TypeToUjoConverter):
    def convert(self):
        return self.type_id


class ListOfTypesToUjo(TypeToUjoConverter):
    def convert(self):
        type_list = UjoList()
        for typ in self.type:
            type_list.append(self.convert_type(typ))
        return type_list


class TypedefToUjo(TypeToUjoConverter):
    def resolve_custom_type(self):
        """ replace the current type with the referenced one,
            but keep current name & doc-string and merge any constraints """
        customtype = deepcopy(self.schema[self.type.type.name])

        # keep original name & doc-string
        for attr in ("name", "doc"):
            setattr(customtype, attr, getattr(self.type, attr))

        # merge constraints
        if any(hasattr(t, "constraint") for t in (customtype, self.type)):
            setattr(customtype, "constraint", getattr(customtype, "constraint", []))
            customtype.constraint += getattr(self.type, "constraint", [])

        self.type = customtype

    def convert(self):
        if isinstance(self.type.type, CustomType):
            self.resolve_custom_type()

        # resolving the CustomType may change our "Typdef"-typ into something else

        if isinstance(self.type, (Record, Object)):
            container = self.convert_type(self.type)
            container.set_value(0, self.type_name, "FixMe")
            # there can be no constraints here, so we can already return this
            return container

        if isinstance(self.type.type, (UjsList, UjsMap)):
            container = self.convert_type(self.type.type)
            container.set_value(0, self.type_name, "FixMe")
        else:
            container = UjoList()
            container.append(self.type_name)
            container.append(self.convert_type(self.type.type))

        container.append(constraints_to_ujo(self.type))
        return container


class ContainerToUjo(TypeToUjoConverter):
    def convert(self):
        container = self.create_container()
        container.append(self.convert_type(self.type.type))
        return container


class RecordToUjo(TypeToUjoConverter):
    def convert(self):
        if getattr(self.type, "extends"):
            container = self.convert_type(self.schema[self.type.extends.name])
            container.set_value(0, self.type_name, "FIXME")
        else:
            container = self.create_container()
            container.append(UjoList())

        fields = container[-1]
        for item in self.type:
            fields.append(self.convert_type(item))
        return container


class ObjectToUjo(TypeToUjoConverter):
    def convert(self):
        if isinstance(self.type.type, CustomType):
            container = self.convert_type(self.type.type)
            container.set_value(0, self.type_name, "FIXME")
        else:
            container = self.create_container()
            container.append(UjoMap())

        items = container[-1]
        for name, item in self.convert_items():
            items[name] = item

        return container

    def convert_items(self):
        for item in self.type:
            yield UjoStringUTF8(item.name), self.convert_item(item)

    def convert_item(self, item):
        if isinstance(item.type, (UjsList, UjsMap, Record, Object, CustomType)):
            container = self.convert_type(item.type)
            container.pop(0)  # remove name
        else:
            container = UjoList()
            # "or self.typ.type[1]" to get the "value-type" defined for the object
            # in case the item itself was defined without any type
            container.append(self.convert_type(item.type or self.type.type.type[1]))

        container.append(constraints_to_ujo(item))
        container.append(UjoBool(item.optional))
        return container


class CustomTypeToUjo(TypeToUjoConverter):
    def convert(self):
        return self.convert_type(self.schema[self.type.name])


def constraints_to_ujo(type_instance):
    # constraints
    converters = {
        UjsNotNull: NotNullToUjo,
        UjsOfLength: LengthToUjo,
        UjsIn: InOrExcludeToUjo,
        UjsNotIn: InOrExcludeToUjo,
    }

    container = UjoList()
    for constraint in type_instance.constraint:
        converter = converters.get(type(constraint))
        container.append(converter(constraint, type_instance.default_to_ujo).convert())
    return container


class ConstraintToUjoConverter(BaseConstraintConverter, ABC):
    def __init__(self, constraint, ujo_default_factory):
        super().__init__(constraint)
        self.ujo_default_factory = ujo_default_factory

    def create_container(self):
        container = UjoList()
        container.append(UjoUInt8(self.constraint.typeid))
        return container

    def range_to_ujo(self, range_):
        return RangeToUjo(range_, self.ujo_default_factory).convert()

    @abstractmethod
    def convert(self):
        pass


class NotNullToUjo(ConstraintToUjoConverter):
    def convert(self):
        container = self.create_container()
        if self.constraint.default is not None:
            container.append(self.ujo_default_factory(self.constraint.default))
        else:
            container.append(UJO_VARIANT_NONE)
        return container


class LengthToUjo(ConstraintToUjoConverter):
    
    def convert(self):
        self.ujo_default_factory = UjoUInt32

        container = self.create_container()
        container.append(self.range_to_ujo(self.constraint.length))
        return container


class InOrExcludeToUjo(ConstraintToUjoConverter):
    def convert(self):
        container = self.create_container()
        values = UjoList()
        container.append(values)
        for value in self.constraint:
            values.append(self.range_to_ujo(value))
        return container


class RangeToUjo(BaseRangeConverter):
    def __init__(self, range_, ujo_default_factory):
        super().__init__(range_)
        self.ujo_default_factory = ujo_default_factory

    def convert(self):
        if isinstance(self.range, (UjsEquals, UjsWord)):
            return self.ujo_default_factory(self.range.value)

        container = UjoList()

        if isinstance(self.range, UjsBetween):
            container.append(self.ujo_default_factory(self.range.low))
            container.append(self.ujo_default_factory(self.range.high))

        elif isinstance(self.range, UjsAbove):
            container.append(self.ujo_default_factory(self.range.value))
            container.append(UJO_VARIANT_NONE)

        elif isinstance(self.range, UjsBelow):
            container.append(UJO_VARIANT_NONE)
            container.append(self.ujo_default_factory(self.range.value))

        return container



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" type definitions """

from UJOSchema.types.container import UjsList

from .basic import UjsBase


class TypedefBase:
    def __init__(self, name):
        self.name = name
        self.doc = None


class Module(TypedefBase):
    def __repr__(self):
        s = "module " + self.name
        if self.doc:
            s = s + "::doc '" + self.doc + "'"
        return s


class Typedef(TypedefBase):
    def __init__(self, name, typ):
        super().__init__(name)
        self.type = typ
        self.constraint = []
        if isinstance(self.type, UjsBase):
            self.default_to_ujo = self.type.default_to_ujo
            self.typeid = self.type.typeid

        if isinstance(self.type, (UjsList, list)):

            def guess_default_type(val):
                types = self.type.type if isinstance(self.type, UjsList) else self.type
                for typ in types:
                    try:
                        return typ.default_to_ujo(val)
                    except (TypeError, ValueError):
                        pass
                raise ValueError(f"No matching type found for default: {val!r}")

            self.default_to_ujo = guess_default_type

    def __repr__(self):
        s = self.name + "=" + repr(self.type) + "::" + repr(self.constraint)
        if self.doc:
            s = s + "::doc '" + self.doc + "'"
        return s



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

"""
Range values with documentation used in constraints like `length` and `in`.
"""

from abc import ABC, abstractmethod


class UjsValueBase(ABC):
    def __init__(self, doc=None):
        self.doc = doc

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def __repr__(self):
        pass


class UjsEquals(UjsValueBase):
    def __init__(self, value, doc=None):
        self.value = value
        super().__init__(doc)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        s = str(self.value) if (not self.doc) else str(self.value) + ":" + self.doc
        return s


class UjsBetween(UjsValueBase):
    def __init__(self, low, high, doc=None):
        super().__init__(doc)
        self.low = low
        self.high = high

    def __str__(self):
        return str(self.low) + " .. " + str(self.high)

    def __repr__(self):
        s = str(self.low) + " .. " + str(self.high)
        if self.doc:
            s = s + ":" + self.doc
        return s


class UjsAbove(UjsValueBase):
    def __init__(self, value, doc=None):
        super().__init__(doc)
        self.value = value

    def __str__(self):
        return ">=" + str(self.value)

    def __repr__(self):
        s = ">=" + str(self.value)
        if self.doc:
            s = s + ":" + self.doc
        return s


class UjsBelow(UjsValueBase):
    def __init__(self, value, doc=None):
        super().__init__(doc)
        self.value = value

    def __str__(self):
        return "<=" + str(self.value)

    def __repr__(self):
        s = "<=" + str(self.value)
        if self.doc:
            s = s + ":" + self.doc
        return s


class UjsWord(UjsValueBase):
    def __init__(self, value, doc=None):
        super().__init__(doc)
        self.value = value

    def __str__(self):
        return "'" + self.value + "'"

    def __repr__(self):
        s = "'" + self.value + "'"
        if self.doc:
            s = s + ":" + self.doc
        return s



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

# pylint: disable=super-init-not-called

""" constraints used in ujo schema """

from abc import ABC, abstractmethod


class UjsConstraint(ABC):
    """ base class for constraints """

    @abstractmethod
    def __repr__(self):
        pass


class UjsNotNull(UjsConstraint):
    typeid = 0x01

    def __init__(self):
        self.default = None

    def __repr__(self):
        s = "not null"
        if self.default is not None:
            s = s + " default " + repr(self.default)
        return s


class UjsOfLength(UjsConstraint):
    typeid = 0x02

    def __init__(self, length):
        self.length = length

    def __repr__(self):
        return "length( " + repr(self.length) + " )"

    def __iter__(self):
        return iter(self.length)


class UjsContains(UjsConstraint):
    typeid = 0x03

    def __init__(self, types):
        self.types = types

    def __repr__(self):
        return "of " + repr(self.types)


class UjsIn(UjsConstraint):
    typeid = 0x04

    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return "in " + repr(self.values)

    def __iter__(self):
        return iter(self.values)


class UjsNotIn(UjsConstraint):
    typeid = 0x05

    def __init__(self, values):
        self.values = values

    def __repr__(self):
        return "not in " + repr(self.values)

    def __iter__(self):
        return iter(self.values)



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" records """

from .typedef import TypedefBase, Typedef


class Record(TypedefBase):
    typeid = 0xE0

    def __init__(self, name):
        super().__init__(name)
        self.items = []
        self.doc = ""
        self.extends = None

    def __repr__(self):
        s = self.name
        if self.extends is not None:
            s = s + " extends " + repr(self.extends)
        s = s + repr(self.items)
        if self.doc:
            s = s + "::doc '" + self.doc + "'"
        return s

    def __iter__(self):
        return iter(self.items)


class RecordItem(Typedef):
    pass



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" basic types """

from ujotypes import (
    UjoInt8,
    UjoInt16,
    UjoInt32,
    UjoInt64,
    UjoUInt8,
    UjoUInt16,
    UjoUInt32,
    UjoUInt64,
    UjoFloat16,
    UjoFloat32,
    UjoFloat64,
    UjoStringUTF8,
    UjoStringC,
    UjoBool,
    UjoTimestamp,
    UJO_VARIANT_NONE,
)


def undefined_ujo_type(*args, **kwargs):  # pylint: disable=unused-argument
    return UJO_VARIANT_NONE


class UjsBase:
    name = "abstract class UjsTypeAtomic"
    typeid = 0xFF
    default_to_ujo = undefined_ujo_type

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class UjsInt8(UjsBase):
    name = "int8"
    typeid = 0x08
    default_to_ujo = UjoInt8


class UjsInt16(UjsBase):
    name = "int16"
    typeid = 0x07
    default_to_ujo = UjoInt16


class UjsInt32(UjsBase):
    name = "int32"
    typeid = 0x06
    default_to_ujo = UjoInt32


class UjsInt64(UjsBase):
    name = "int64"
    typeid = 0x05
    default_to_ujo = UjoInt64


class UjsUInt8(UjsBase):
    name = "uint8"
    typeid = 0x0C
    default_to_ujo = UjoUInt8


class UjsUInt16(UjsBase):
    name = "uint16"
    typeid = 0x0B
    default_to_ujo = UjoUInt16


class UjsUInt32(UjsBase):
    name = "uint32"
    typeid = 0x0A
    default_to_ujo = UjoUInt32


class UjsUint64(UjsBase):
    name = "uint64"
    typeid = 0x09
    default_to_ujo = UjoUInt64


class UjsFloat16(UjsBase):
    name = "float16"
    typeid = 0x03
    default_to_ujo = UjoFloat16


class UjsFloat32(UjsBase):
    name = "float32"
    typeid = 0x02
    default_to_ujo = UjoFloat32


class UjsFloat64(UjsBase):
    name = "float64"
    typeid = 0x01
    default_to_ujo = UjoFloat64


class UjsUtf8String(UjsBase):
    name = "string"
    typeid = 0x04
    default_to_ujo = UjoStringUTF8


class UjsCString(UjsBase):
    name = "cstring"
    typeid = 0x04
    default_to_ujo = UjoStringC


class UjsBinary(UjsBase):
    name = "binary"
    typeid = 0x0D
    default_to_ujo = undefined_ujo_type  # UjoBinary ??


class UjsBool(UjsBase):
    name = "bool"
    typeid = 0x0E
    default_to_ujo = UjoBool


class UjsDateTime(UjsBase):
    name = "datetime"
    typeid = 0x10
    default_to_ujo = undefined_ujo_type  # UjoDateTime?


class UjsDate(UjsBase):
    name = "date"
    typeid = 0x11
    default_to_ujo = undefined_ujo_type  # UjoDate?


class UjsTime(UjsBase):
    name = "time"
    typeid = 0x12
    default_to_ujo = undefined_ujo_type  # UjoTime?


class UjsTimestamp(UjsBase):
    name = "timestamp"
    typeid = 0x13
    default_to_ujo = UjoTimestamp


class UjsVariant(UjsBase):
    typeid = 0xAA
    name = "variant"
    default_to_ujo = undefined_ujo_type


# custom type
class CustomType(UjsBase):
    def __init__(self, name):
        self.name = name



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#


""" serialization of a schema into JSON """

import json
from abc import ABC, abstractmethod

from UJOSchema.types import Module, Object, Record
from UJOSchema.constraints import UjsNotNull, UjsIn, UjsNotIn, UjsOfLength
from UJOSchema.ranges import UjsAbove, UjsBelow, UjsBetween

from .base import BaseSchemaConverter, BaseTypeConverter


def schema_to_json(schema, target=None):
    return SchemaToJSON(schema).convert(target)


class SchemaToJSON(BaseSchemaConverter):
    def __init__(self, schema):
        super().__init__(schema)
        self.type_converter = TypeToJSONConverter(self.schema, None)
        self.convert_type = self.type_converter.convert_type

    def convert(self, target=None):  # pylint: disable=arguments-differ

        converted = [
            self.convert_type(typ) for typ in self.schema.values() if not isinstance(typ, Module)
        ]

        converted = merge_extends(converted)

        if target:
            converted = get_target(converted, target)
        return json.dumps(list(converted.values()), indent=4)


class TypeToJSONConverter(BaseTypeConverter):
    def __init__(self, schema, typ, optional=False):
        super().__init__(schema, typ)
        self.__elements = None
        self.optional = optional

    def convert(self):
        return self.convert_type(self.type)

    def convert_type(self, typ, optional=False):
        default_converter = BasicTypeToJSON
        converters = {Record: ContainerToJSON, Object: ContainerToJSON}

        converter = converters.get(type(typ)) or default_converter
        return converter(self.schema, typ, optional).convert()

    @property
    def name(self):
        return "" if not self.type else self.type.name

    @property
    def type_name(self):
        return ""

    @property
    def default(self):
        return None

    @property
    def doc(self):
        return "" if not self.type else self.type.doc or ""

    @property
    def list_type(self):
        if self.type_name == "list":
            return self.type.type.type.name
        return None

    @property
    def key_value_types(self):
        if self.type_name == "map":
            return [typ.name for typ in self.type.type.type]
        return None

    @property
    def constraints(self):
        if self.type and hasattr(self.type, "constraint"):
            return get_constraints(self.type.constraint)
        return None

    @property
    def elements(self):
        return self.__elements

    @elements.setter
    def elements(self, values):
        self.__elements = values

    def to_dict(self):
        json_elements = {"name": self.name, "type": self.type_name, "doc": self.doc}
        if self.constraints:
            json_elements["constraints"] = self.constraints
        if self.list_type:
            json_elements["list_type"] = self.list_type
        if self.key_value_types:
            json_elements["key_value_types"] = self.key_value_types
        if self.elements:
            json_elements["elements"] = self.elements
        if self.optional:
            json_elements["optional"] = self.optional
        if self.default:
            json_elements["default"] = self.default
        return json_elements


class BasicTypeToJSON(TypeToJSONConverter):
    @property
    def default(self):
        if self.constraints:
            notNull = (
                next((c for c in self.type.constraint if isinstance(c, UjsNotNull)), None)
                or UjsNotNull()
            )
            return notNull.default
        return None

    @property
    def type_name(self):
        if hasattr(self.type.type, "name"):
            return self.type.type.name

        return [typ.name for typ in self.type.type]

    def convert(self):
        return self.to_dict()


class ContainerToJSON(TypeToJSONConverter):
    @property
    def type_name(self):
        if getattr(self.type, "extends"):
            return "extends_" + self.type.extends.name

        if isinstance(self.type, Record):
            return "record"

        if isinstance(self.type, Object):
            return "object"
        return ""

    def convert(self):
        elem = []
        for item in self.type:
            optional = item.optional if hasattr(item, "optional") else False
            if not item.type:
                item.type = self.type.type.type[1]
            elem.append(self.convert_type(item, optional))

        self.elements = elem
        return self.to_dict()


def get_constraints(typ_constraints):
    if not typ_constraints:
        return None

    converters = {
        UjsNotNull: NotNullConstraintToJSON,
        UjsOfLength: LengthConstraintToJSON,
        UjsIn: InConstraintToJSON,
        UjsNotIn: NotInConstraintToJSON,
    }
    constraints = []
    for constraint in typ_constraints:
        converter = converters.get(type(constraint))
        constraints.append(converter(constraint).convert())
    return constraints


class ConstraintToJSONConverter(ABC):
    def __init__(self, constraint):
        self.constraint = constraint

    @abstractmethod
    def convert(self):
        pass


class InConstraintToJSON(ConstraintToJSONConverter):
    def convert(self):
        default_converter = InConverter
        converters = {UjsBelow: BelowConverter, UjsAbove: AboveConverter}
        converter = converters.get(type(self.constraint.values[0])) or default_converter
        return {"in": converter(self.constraint).convert()}


class AboveConverter(ConstraintToJSONConverter):
    def convert(self):
        above = str(self.constraint.values).replace(">=", "") + ".."
        return above.split("..")


class InConverter(ConstraintToJSONConverter):
    def convert(self):
        in_constraints = []
        for con in self.constraint.values:
            if isinstance(con, UjsBetween):
                return [str(c) for c in str(con).split(" ")]
            in_constraints.append(str(con))
        return in_constraints


class BelowConverter(ConstraintToJSONConverter):
    def convert(self):
        return str(self.constraint.values).replace("<=", ".. ").split(" ")


class NotInConstraintToJSON(ConstraintToJSONConverter):
    def convert(self):
        default_converter = NotInConverter
        converters = {UjsBelow: BelowConverter, UjsAbove: NotInAboveConverter}
        converter = converters.get(type(self.constraint.values)) or default_converter
        return {"notIn": converter(self.constraint).convert()}


class NotInConverter(ConstraintToJSONConverter):
    def convert(self):
        if isinstance(self.constraint.values, UjsBetween):
            return [str(c) for c in str(self.constraint.values).split(" ")]
        return []


class NotInAboveConverter(ConstraintToJSONConverter):
    def convert(self):
        return [str(self.constraint.values).replace(">=", ""), ".."]


class LengthConstraintToJSON(ConstraintToJSONConverter):
    def convert(self):
        default_converter = LengthConverter
        converters = {UjsBelow: LengthBelowConverter, UjsAbove: LengthAboveConverter}
        converter = converters.get(type(self.constraint.length)) or default_converter
        return {"length": [converter(self.constraint).convert()]}


class LengthBelowConverter(ConstraintToJSONConverter):
    def convert(self):
        return str(self.constraint.length).replace("<=", "..")


class LengthAboveConverter(ConstraintToJSONConverter):
    def convert(self):
        return str(self.constraint.length).replace(">=", "") + ".."


class LengthConverter(ConstraintToJSONConverter):
    def convert(self):
        return str(self.constraint.length)


class NotNullConstraintToJSON(ConstraintToJSONConverter):
    def convert(self):
        return {"notNull": True}


def merge_extends(type_list):
    """merge extended type definitions into their parent types"""
    extensions = {
        typ.get("type").replace("extends_", ""): typ
        for typ in type_list
        if "extends" in typ.get("type")
    }
    extended = {typ.get("name"): typ for typ in type_list}
    for type_name, typ in extensions.items():
        extended[typ.get("name")]["type"] = extended[type_name].get("type")
        extended[typ.get("name")]["elements"] = (
            extended[type_name].get("elements") + extended[typ.get("name")]["elements"]
        )
    return extended


def get_target(converted, target):
    """get only target type and its custom types"""
    target_types = {target: converted[target]}

    if isinstance(converted[target].get("type"), list):
        return target_types

    if converted[target].get("type") in converted:
        custom = get_target(converted, converted[target].get("type"))
        target_types.update(custom)

    if "elements" in converted[target]:
        for element in converted[target].get("elements"):
            if element.get("type") in converted:
                custom_element = get_target(converted, element.get("type"))
                target_types.update(custom_element)

    return target_types



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" base classes for converters """

from abc import ABC, abstractmethod

# pylint: disable=abstract-method
from typing import Optional


class BaseConverter(ABC):
    @abstractmethod
    def convert(self):
        pass


class BaseSchemaConverter(BaseConverter, ABC):
    def __init__(self, schema, target: Optional[str] = None):
        self.schema = schema
        self.target = target


class BaseTypeConverter(BaseConverter, ABC):
    def __init__(self, schema, typ):
        self.schema = schema
        self.type = typ


class BaseConstraintConverter(BaseConverter, ABC):
    def __init__(self, constraint):
        self.constraint = constraint


class BaseRangeConverter(BaseConverter, ABC):
    def __init__(self, range_):
        self.range = range_



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" conversion of a schema into a an empty UjoType instance """

from datetime import datetime
from typing import Union, Sequence

from ujotypes import UjoMap, UJO_VARIANT_NONE, UjoList

from UJOSchema.constraints import UjsNotNull
from UJOSchema.schema import UjsSchema
from UJOSchema.types import UjsUtf8String, Typedef, UjsList, UjsMap, CustomType, Record, Object
from UJOSchema.converter.base import BaseSchemaConverter, BaseTypeConverter


UJO_EMPTY_DEFAULTS = {
    "bool": False,
    "int8": 0,
    "int16": 0,
    "int32": 0,
    "int64": 0,
    "uint8": 0,
    "uint16": 0,
    "uint32": 0,
    "uint64": 0,
    "float16": 0.0,
    "float32": 0.0,
    "float64": 0.0,
    "date": "nyi",
    "time": "nyi",
    "datetime": "nyi",
    "timestamp": datetime(2000, 1, 1),
    "string": "",
    "cstring": "",
    "binary": "nyi",  # bytes(),
    "variant": None,
}


def schema_to_type(schema: Union[str, UjsSchema], target: str):
    if isinstance(schema, str):
        schema = UjsSchema.from_str(schema)

    return SchemaToType(schema, target).convert()


class SchemaToType(BaseSchemaConverter):
    def __init__(self, schema, target):
        super().__init__(schema, target)

    def convert(self):
        return TypeToUjoTypeConverter(self.schema, self.schema[self.target]).convert()


class TypeToUjoTypeConverter(BaseTypeConverter):
    def convert(self):
        default_converter = BasicTypeToUjo
        converters = {
            CustomType: CustomTypeToUjo,
            UjsList: ContainerToUjo,
            UjsMap: ContainerToUjo,
            Record: RecordToUjo,
            Object: ObjectToUjo,
        }
        if isinstance(self.type, Typedef):
            converter = converters.get(type(self.type.type)) or default_converter
        else:
            converter = converters.get(type(self.type)) or default_converter

        return converter(self.schema, self.type).convert()


class BasicTypeToUjo(BaseTypeConverter):
    def convert(self):
        # TO_DO: consider: not null: default and constraints in / not in (range)

        default_type = self.type.type
        if isinstance(self.type.type, Sequence):
            default_type = self.type.type[0]

        default = get_default_value(self.type.constraint, default_type.name)
        return self.type.default_to_ujo(default)


class CustomTypeToUjo(BaseTypeConverter):
    def convert(self):
        return schema_to_type(self.schema, target=self.type.type.name)


class ContainerToUjo(BaseTypeConverter):
    def convert(self):
        return self.type.type.default_to_ujo()


class RecordToUjo(BaseTypeConverter):
    def convert(self):
        if self.type.extends:
            result_list = schema_to_type(self.schema, self.type.extends.name)
        else:
            result_list = UjoList()

        for item in self.type:
            value = TypeToUjoTypeConverter(self.schema, item).convert()
            result_list.append(value)
        return result_list


class ObjectToUjo(BaseTypeConverter):
    def convert(self):
        if self.type.extends:
            result_map = schema_to_type(self.schema, target=self.type.extends.name)
            self.type.type = self.schema[self.type.extends.name].type
        else:
            result_map = UjoMap()

        class Undefined:
            name = "undefined"
            default_to_ujo = lambda *args: UJO_VARIANT_NONE  # noqa

        try:
            key_types, value_types = self.type.type.type
        except (AttributeError, TypeError):
            # in case the definition was not explict // without <key -> value> definition
            key_types, value_types = UjsUtf8String, Undefined

        if not isinstance(key_types, Sequence):
            key_types = [key_types]

        for item in self.type.items:
            item_type = item.type or value_types
            if isinstance(item_type, Sequence):
                # let's simply take the first item to instantiate a value
                item_type = item_type[0]

            if isinstance(item.type, CustomType):
                value = schema_to_type(self.schema, target=item.type.name)
            else:
                default = get_default_value(item.constraint, item_type.name)
                value = item_type.default_to_ujo(default)

            for key_type in key_types:
                # try to find a matching key type for the field name
                try:
                    result_map[key_type.default_to_ujo(item.name)] = value
                except (TypeError, ValueError):
                    continue
                else:
                    break
            else:
                raise TypeError(f"No key type matching {item.name} (available: {key_types})")

        return result_map


def get_default_value(constraints, type_name):
    not_null = next((c for c in constraints if isinstance(c, UjsNotNull)), None) or UjsNotNull()
    if isinstance(not_null.default, int) and "float" in type_name:
        not_null.default = float(not_null.default)

    return not_null.default or UJO_EMPTY_DEFAULTS.get(type_name)



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

# flake8: noqa

from .types import *
from .ranges import *
from .constraints import *
from .parser import UjsParser
from .transformer import UjsTransformer
from .schema import schema_from_str
from .converter import *

# pylint: disable=wrong-import-order
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" parsing of schema files and strings """

import os.path
from lark import Lark


def to_filename_tuple(filename_or_dir, extensions):
    """
    Get all filenames from a directory as a tuple.
    """
    path = filename_or_dir
    if not os.path.isdir(path):
        files = (path,)
    else:
        files = (
            os.path.join(path, f) for f in os.listdir(path) if f.endswith(extensions)
        )
    return files


class UjsParser:
    """
    Create a parse tree for ujs files
    """

    def __init__(
        self,
        grammar_file_or_folder=os.path.join(os.path.dirname(__file__), "grammar"),
        start="start",
    ):
        grammar = self.get_grammar(grammar_file_or_folder)
        self._parser = Lark(grammar, parser="lalr", start=start)

    def parse_file(self, filename):
        """
        Load code from a textfile and return a parse tree.

        Parameter:
            filename : [string] path and filename of the source

        Return:
            [lark.tree] parse tree
        """
        with open(filename) as f:
            ast = self._parser.parse(f.read())
        return ast

    def parse_string(self, text):
        """
        Load code from a textfile and return a parse tree.

        Parameter:
            filename : [text] source

        Return:
            [lark.tree] parse tree
        """
        ast = self._parser.parse(text)
        return ast

    @property
    def parser(self):
        return self.parser

    @parser.setter
    def parser(self, new_parser):
        assert isinstance(new_parser, Lark), "Only Lark objects supported"
        self._parser = new_parser

    @staticmethod
    def get_grammar(file_or_folder):
        grammar = ""
        for filename in to_filename_tuple(
            file_or_folder, extensions=(".g", ".grammar", ".lark")
        ):
            with open(filename) as f:
                grammar += f.read()
        return grammar



"""UJOSchema.

Usage:
  UJOSchema (-h | --help)
  UJOSchema into markdown <source> [-d FOLDER] [-e EXTENSION]
  UJOSchema into binary <source> [-d FOLDER] [-e EXTENSION] [-t TARGET]
  UJOSchema into json <source> [-d FOLDER] [-e EXTENSION] [-t TARGET]

Options:
  -h --help  Show this screen.

  into markdown <source>
      convert given SOURCE (file or folder) to markdown

  into binary <source>
      convert given SOURCE (file or folder) to binary
  
  into json <source>
      convert given SOURCE (file or folder) to JSON

  -d FOLDER --destination FOLDER
      destination folder for generated files [default: .]

  -e EXTENSION --extension EXTENSION
      if SOURCE is a folder, only process files with the specified extension [default: .ujs]

  -t TARGET --target TARGET
      name of a definition target. other types in the schema will be ignored, unless nested.
"""

import sys
import logging
from enum import Enum
from pathlib import Path

from docopt import docopt
from lark.exceptions import LarkError

from UJOSchema import schema_from_str, schema_to_markdown, schema_to_binary, schema_to_json


class ConversionTargetFormats(Enum):
    MARKDOWN = "markdown"
    BINARY = "binary"
    JSON = "json"


def main():
    arguments = docopt(__doc__)

    if arguments.get("into"):
        if arguments.get("markdown"):
            target_format = ConversionTargetFormats.MARKDOWN
        elif arguments.get("binary"):
            target_format = ConversionTargetFormats.BINARY
        elif arguments.get("json"):
            target_format = ConversionTargetFormats.JSON
        else:
            return  # docopt should prevent us from reaching this anyway

        convert(
            source=arguments["<source>"],
            destination=arguments["--destination"],
            extension=arguments["--extension"],
            target_format=target_format,
            target_type=arguments.get("--target"),
        )


def convert(source, destination, extension, target_format, target_type=None):
    try:
        destination = Path(destination).resolve()
        destination.mkdir(parents=True, exist_ok=True)
    except OSError as error:
        log.error(f"Failed to create output directory: {error}")
        return

    for schema_file in get_schema_files(source, extension):
        log.info(f"converting: {schema_file} ...")

        schema = schema_from_file(schema_file)
        if not schema:
            continue

        if target_format is ConversionTargetFormats.MARKDOWN:
            md_file = destination / (schema_file.stem + ".md")
            md = schema_to_markdown(schema)
            write_to_file(destination=md_file, mode="w+", content=md)
        elif target_format is ConversionTargetFormats.BINARY:
            bin_file = destination / ((target_type or schema_file.stem) + ".bin")
            bin_data = schema_to_binary(schema, target_type)
            write_to_file(destination=bin_file, mode="wb+", content=bin_data)
        elif target_format is ConversionTargetFormats.JSON:
            json_file = destination / ((target_type or schema_file.stem) + ".json")
            json_data = schema_to_json(schema, target_type)
            write_to_file(destination=json_file, mode="w+", content=json_data)


def get_schema_files(source, extension):
    try:
        source_path = Path(source).resolve()
    except OSError as error:
        log.error(f"Failed to process source: {error}")
        return []

    if source_path.is_dir():
        schema_files = source_path.glob("*." + extension.split(".")[-1])
    else:
        schema_files = [source_path]
    return schema_files


def schema_from_file(schema_file):
    try:
        with open(schema_file) as sf:
            schema = schema_from_str(sf.read())
    except OSError as error:
        log.error(f"Failed to read {schema_file}: {error}")
    except LarkError as error:
        log.error(f"Failed to parse {schema_file}: {error}")
    else:
        return schema


def write_to_file(destination, mode, content):

    try:
        with open(destination, mode) as md_file:
            md_file.write(content)
    except OSError:
        log.error(f"Failed to write to {destination}")
    else:
        log.info(f"... done: {destination}")


if __name__ == "__main__":

    formatter = logging.Formatter("%(levelname)s: %(message)s")

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    log = logging.getLogger(Path(__file__).parent.stem)
    log.addHandler(handler)
    # log.setLevel(logging.INFO)
    log.setLevel(logging.DEBUG)

    main()



from collections.abc import MutableMapping

from .parser import UjsParser
from .transformer import UjsTransformer


def schema_from_str(schema_str):
    """
    create UJOSchema type definitions from a string representation

    parameter:
        schema_str: [string] a schema definition string

    return:
        [dictionary] type definition objects identified by names
    """
    return UjsSchema.from_str(schema_str)


class UjsSchema(MutableMapping):
    def __init__(self, schema: dict):
        self._schema = schema

    def __getitem__(self, key):
        return self._schema[key]

    def __setitem__(self, key, value):
        self._schema[key] = value

    def __delitem__(self, key):
        del self._schema[key]

    def __iter__(self):
        return iter(self._schema)

    def __len__(self):
        return len(self._schema)

    def __repr__(self):
        return self._schema.__repr__()

    @classmethod
    def from_str(cls, schema_str: str):
        parser = UjsParser(start="begin")
        parse_tree = parser.parse_string(schema_str)

        transformer = UjsTransformer()
        schema = transformer.transform(parse_tree)

        return cls(schema)



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import setuptools

with open("doc/markdown/ujo-schema.md", "r") as fh:
    long_description = fh.read()

version = '0.3'
try:
    # extend baseversion with buildnumber as microversion forming
    # a version scheme of release.minor.micro
    version = '%s.%s' % (version, os.environ['CI_PIPELINE_IID'])
except BaseException:
    pass

setuptools.setup(
    name="UJOSchema",
    version=version,
    author="Maik Wojcieszak",
    author_email="mw@wobew-systems.com",
    description="UJO Schema is an easy to read and easy"
                "to write language to define UJO data structures",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.industrial-devops.org",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={
        "UJOSchema": ["grammar/*.lark"],
    },
    data_files=[
        ("", ["LICENSE"])
    ],
    install_requires=[
        "lark-parser",
        "docopt",
        "ujotypes",
    ],
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" transforming the parsed AST """

import logging

from lark import Transformer

from .ranges import UjsWord, UjsAbove, UjsBetween, UjsBelow, UjsEquals
from .constraints import UjsNotNull, UjsConstraint, UjsOfLength, UjsIn, UjsContains, UjsNotIn
from .types import (
    UjsInt8,
    UjsInt16,
    UjsInt32,
    UjsInt64,
    UjsUInt8,
    UjsUInt16,
    UjsUInt32,
    UjsUint64,
    UjsFloat16,
    UjsFloat32,
    UjsFloat64,
    UjsUtf8String,
    UjsCString,
    UjsBinary,
    UjsBool,
    UjsDate,
    UjsTime,
    UjsDateTime,
    UjsTimestamp,
    UjsVariant,
    UjsList,
    UjsMap,
    Typedef,
    Module,
    CustomType,
    Object,
    ObjectItem,
    Record,
    RecordItem,
)

log = logging.getLogger(__name__)


# transformer
class UjsTransformer(Transformer):
    # pylint: disable=no-self-use, unused-argument, too-many-public-methods

    def __init__(self):
        super().__init__(self)
        self.parse_tree = None

    def not_null(self, matches):
        log.debug("not null")
        log.debug(matches)
        result = UjsNotNull()
        if matches:
            result.default = matches[0]
        return result

    def default(self, matches):
        log.debug("default")
        log.debug(matches)
        return matches[0]

    def documentation(self, matches):
        # log.debug("doc")
        # log.debug(matches)

        if matches[0].type == "LONG_STRING":
            return matches[0].value[3:-3]
        if matches[0].type == "STRING":
            return matches[0].value[1:-1]

        raise Exception("invalid documentation string type")

    def int8(self, matches):
        return UjsInt8()

    def int16(self, matches):
        return UjsInt16()

    def int32(self, matches):
        return UjsInt32()

    def int64(self, matches):
        return UjsInt64()

    def uint8(self, matches):
        return UjsUInt8()

    def uint16(self, matches):
        return UjsUInt16()

    def uint32(self, matches):
        return UjsUInt32()

    def uint64(self, matches):
        return UjsUint64()

    def float16(self, matches):
        return UjsFloat16()

    def float32(self, matches):
        return UjsFloat32()

    def float64(self, matches):
        return UjsFloat64()

    def utf8string(self, matches):
        return UjsUtf8String()

    def string(self, matches):
        return UjsUtf8String()

    def cstring(self, matches):
        return UjsCString()

    def binary(self, matches):
        return UjsBinary()

    def bool(self, matches):
        return UjsBool()

    def date(self, matches):
        return UjsDate()

    def time(self, matches):
        return UjsTime()

    def datetime(self, matches):
        return UjsDateTime()

    def timestamp(self, matches):
        return UjsTimestamp()

    def list(self, matches):
        log.debug("list")
        log.debug(matches)
        return UjsList(matches[0])

    def variant(self, matches):
        return UjsVariant()

    def identifier(self, matches):
        log.debug("identifier")
        log.debug(matches)
        return matches[0].value

    def of_length(self, matches):
        log.debug("of_length")
        log.debug(matches)

        return UjsOfLength(matches[0])

    def primitive_type(self, matches):
        log.debug("primitive_type")
        log.debug(matches)

        result = Typedef(matches[0], matches[1])
        for element in matches[2:]:
            if isinstance(element, UjsConstraint):
                result.constraint.append(element)

        # documentation found
        if isinstance(matches[-1], str):
            result.doc = matches[-1]

        del matches[2:]
        return (matches[0], result)

    def custom_type(self, matches):
        log.debug("custom_type")
        log.debug(matches)

        return CustomType(matches[0].value)

    def value(self, matches):
        # log.debug("value")
        # log.debug(matches)
        try:
            number = int(matches[0].value)
        except ValueError:
            number = float(matches[0].value)
        return number

    def words(self, matches):
        # log.debug("words")
        # log.debug(matches)

        return matches

    def word(self, matches):
        # log.debug("word")
        # log.debug(matches)

        if len(matches) > 1:
            return UjsWord(matches[0], matches[1])

        return UjsWord(matches[0])

    def chars(self, matches):
        # log.debug("chars")
        # log.debug(matches)

        return matches[0].value[1:-1]

    def value_set(self, matches):
        log.debug("value_set")
        log.debug(matches)

        if isinstance(matches[0], list):
            return UjsIn(matches[0])

        return UjsIn(matches)

    def not_in(self, matches):
        log.debug("value_set")
        log.debug(matches)

        return UjsNotIn(matches[0])

    def range(self, matches):
        log.debug("range")
        log.debug(matches)

        return matches

    def above(self, matches):
        # log.debug("above")
        # log.debug(matches)

        if len(matches) > 1:
            return UjsAbove(matches[0], matches[1])

        return UjsAbove(matches[0])

    def between(self, matches):
        # log.debug("between")
        # log.debug(matches)

        if len(matches) == 3:
            # documentation found
            return UjsBetween(matches[0], matches[1], matches[2])

        return UjsBetween(matches[0], matches[1])

    def below(self, matches):
        # log.debug("below")
        # log.debug(matches)

        if len(matches) > 1:
            return UjsBelow(matches[0], matches[1])

        return UjsBelow(matches[0])

    def equals(self, matches):
        log.debug("equals")
        log.debug(matches)

        if len(matches) > 1:
            return UjsEquals(matches[0], matches[1])

        return UjsEquals(matches[0])

    def contains(self, matches):
        log.debug("contains")
        log.debug(matches)

        return UjsContains(matches)

    def item(self, matches):
        log.debug("item")
        log.debug(matches)

        result = RecordItem(matches[0], matches[1])
        for element in matches[2:]:
            if isinstance(element, UjsConstraint):
                result.constraint.append(element)

        # documentation found
        if isinstance(matches[-1], str):
            result.doc = matches[-1]

        return result

    def record(self, matches):
        log.debug("record")
        log.debug(matches)

        result = Record(matches[0])

        for element in matches[1:]:
            if isinstance(element, RecordItem):
                result.items.append(element)
            elif isinstance(element, CustomType):
                result.extends = element

        # documentation found
        if isinstance(matches[-1], str):
            result.doc = matches[-1]

        return (matches[0], result)

    def object(self, matches):
        log.debug("object")
        log.debug(matches)

        result = Object(matches[0], matches[1], [])
        for element in matches[1:]:
            if isinstance(element, ObjectItem):
                result.items.append(element)
            elif isinstance(element, CustomType):
                result.extends = element

        # documentation found
        if isinstance(matches[-1], str):
            result.doc = matches[-1]

        return matches[0], result

    def object_item(self, matches):
        log.debug("object_item")
        log.debug(matches)

        typ = next((m for m in matches[1:] if not isinstance(m, str)), None)

        result = ObjectItem(str(matches[0]), typ)

        for element in matches:
            if isinstance(element, UjsConstraint):
                result.constraint.append(element)
            elif isinstance(element, str) and element == "*?*":
                result.optional = True
            elif isinstance(element, str):
                result.doc = element

        return result

    def optional(self, matches):
        log.debug("multi_type")
        log.debug(matches)

        return "*?*"

    def multi_type(self, matches):
        log.debug("multi_type")
        log.debug(matches)

        return matches

    def map_type(self, matches):
        log.debug("map_type")
        log.debug(matches)

        return UjsMap(tuple(matches))

    def extending(self, matches):
        log.debug("extending")
        log.debug(matches)

        return matches[0]

    def module(self, matches):
        log.debug("extending")
        log.debug(matches)

        result = Module(matches[0])

        if len(matches) > 1:
            result.doc = matches[1]

        return (matches[0], result)

    def begin(self, matches):
        log.debug("begin")
        log.debug(matches)

        return dict(matches)



#
# Copyright (c) 2018-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" conversion of a schema into a markdown representation """
import logging
from string import Template
from textwrap import dedent

from UJOSchema.constraints import UjsIn, UjsNotIn
from .base import BaseSchemaConverter

log = logging.getLogger(__name__)


def schema_to_markdown(schema):
    return str(SchemaToMarkdown(schema))


class SchemaToMarkdown(BaseSchemaConverter):
    def convert(self):
        result = ""
        for _, value in self.schema.items():
            result += self.class2string(value)
        return result

    def __str__(self):
        return self.convert()

    def class2string(self, obj):
        type_name = type(obj).__name__.lower()
        try:
            converter_method = getattr(self, type_name)
        except AttributeError:
            log.warning(f"Possibly missing: {type_name}")
        else:
            return converter_method(obj)

        return str(obj)

    def typedef(self, obj):
        s = Template(
            dedent(
                """\
                ## $identifier
    
                $doc
    
                | Identifier  | Type  | Constraints  |
                |-------------|-------|--------------|
                | $identifier | $type | $constraints |
                """
            )
        )

        constraints = ""
        range_tabs = []
        for element in obj.constraint:
            constraints = constraints + "<br>" if constraints != "" else ""
            if isinstance(element, UjsIn):
                c = "in [Values](#" + obj.name + "-in-values)"
                range_tabs.append(element)
            elif isinstance(element, UjsNotIn):
                c = "not in [Values](#" + obj.name + "-not-in-values)"
                range_tabs.append(element)
            else:
                c = self.class2string(element)
            constraints = constraints + c

        d = {
            "identifier": obj.name,
            "type": str(obj.type),
            "doc": "" if not obj.doc else obj.doc,
            "constraints": constraints,
        }

        s = s.substitute(d)

        for range_tab in range_tabs:

            iv = Template(
                dedent(
                    """\
                    
                    ### $headline
    
                    $values
                    """
                )
            )
            if isinstance(range_tab, UjsIn):
                d = {
                    "headline": obj.name + " in Values",
                    "values": self.class2string(range_tab),
                }
            elif isinstance(range_tab, UjsNotIn):
                d = {
                    "headline": obj.name + " not in Values",
                    "values": self.class2string(range_tab),
                }
            else:
                raise TypeError("invalid constraint")

            iv = iv.substitute(d)
            s = s + iv

        return s

    def record(self, record):  # pylint: disable=no-self-use
        md = Template(
            dedent(
                """\
                ## $identifier

                $doc

                Items:

                $items
                """
            )
        )

        d = {"identifier": record.name, "doc": record.doc}

        items = [["Name", "Type", "Doc", "Constraints"]]
        items += [
            [item.name, item.type, item.doc or "", item.constraint or ""]
            for item in record.items
        ]

        d["items"] = to_markdown_table(items)
        return md.substitute(d)

    def object(self, uobject):  # pylint: disable=no-self-use
        md = Template(
            dedent(
                """\
                ## $identifier

                $doc

                Items:

                $items
                """
            )
        )

        d = {"identifier": uobject.name, "doc": uobject.doc or ""}

        items = [["Name", "Type", "Doc", "Constraints"]]
        items += [
            [item.name, item.type or "", item.doc or "", item.constraint or ""]
            for item in uobject.items
        ]

        d["items"] = to_markdown_table(items)
        return md.substitute(d)

    def ujsin(self, obj):
        s = dedent(
            """\
            | Value | Description |
            |-------|-------------|
            """
        )

        for value in obj.values:
            doc = value.doc if value.doc is not None else ""
            s = s + "| " + self.class2string(value) + " | " + doc + " |\n"
        return s

    def ujsnotin(self, obj):
        return self.ujsin(obj)


def to_markdown_table(rows, header=None):
    if header:
        rows = header, *rows

    rows = [[str(val) for val in row] for row in rows]
    columns = zip(*rows)
    column_widths = [max(len(val) for val in col) for col in columns]

    rows.insert(1, ["-" * w for w in column_widths])

    md = ""
    for row in rows:
        justified = (
            value.ljust(column_widths[col_idx]) for col_idx, value in enumerate(row)
        )
        md += f"| {' | '.join(justified)} |  \n"

    return md



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" utility functions """

import logging

log = logging.getLogger(__name__)


def print_pretty_bytes(value):
    chunk_size = 16
    chunks = (value[i: i + chunk_size] for i in range(0, len(value), chunk_size))

    for chunk in chunks:
        hex_values = ""
        char_values = ""

        for val in chunk:
            hex_values += f"{val:02x} "
            char_values += "." if val < 32 else chr(val)

        log.debug(hex_values + char_values)



#
# Copyright (c) 2020-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

# flake8: noqa

from .basic import (
    UjsBase,
    UjsBinary,
    UjsBool,
    UjsVariant,
    UjsCString,
    UjsUtf8String,
    UjsDate,
    UjsDateTime,
    UjsTime,
    UjsTimestamp,
    UjsFloat16,
    UjsFloat32,
    UjsFloat64,
    UjsInt8,
    UjsInt16,
    UjsInt32,
    UjsInt64,
    UjsUInt8,
    UjsUInt16,
    UjsUInt32,
    UjsUint64,
    CustomType,
)
from .container import UjsList, UjsMap
from .typedef import Typedef, TypedefBase, Module
from .record import Record, RecordItem
from .object import Object, ObjectItem



#
# Copyright (c) 2019-present, wobe-systems GmbH
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

""" objects """

from .typedef import TypedefBase, Typedef


class Object(TypedefBase):
    typeid = 0xE1

    def __init__(self, name, typ, items):
        super().__init__(name)
        self.type = typ
        self.items = items
        self.extends = None

    def __str__(self):
        s = self.name + " object of " + repr(self.type) + "{" + str(self.items) + "}"
        if self.extends is not None:
            s = s + " extends " + repr(self.extends)
        return s

    def __repr__(self):
        s = self.name + " object of " + repr(self.type) + "{" + str(self.items) + "}"
        if self.extends is not None:
            s = s + " extends " + repr(self.extends)
        return s

    def __iter__(self):
        return iter(self.items)


class ObjectItem(Typedef):
    def __init__(self, key, typ):
        super().__init__(str(key), typ)
        self.key = key
        self.optional = False

    def __repr__(self):
        if self.optional:
            s = "?"
        else:
            s = ""
        s = s + repr(self.key) + "->" + repr(self.type) + "::" + repr(self.constraint)
        if self.doc:
            s = s + "::doc '" + self.doc + "'"
        return s
